import unittest, datetime
from binhistory import BinLogEntry, exceptions, defaults

EXAMPLE_STRING = "Mon Mar 10 17:32:54  Computer: zMichael        User: poop           "
EXAMPLE_DATE   = datetime.datetime(year=2025, month=3, day=10, hour=17, minute=32, second=54)
EXAMPLE_ENTRY  = BinLogEntry(timestamp=EXAMPLE_DATE, computer="zMichael", user="poop")

class TestBinLogEntry(unittest.TestCase):

	def test_defaults(self):

		timestamp = datetime.datetime.now()
		self.assertEqual(
			BinLogEntry(timestamp=timestamp),
			BinLogEntry(timestamp=timestamp, user=defaults.DEFAULT_USER, computer=defaults.DEFAULT_COMPUTER)
		)

	def test_parse_string_date(self):

		log_entry = BinLogEntry.from_string(EXAMPLE_STRING, max_year=2025)

		self.assertEqual(log_entry.computer, "zMichael")
		self.assertEqual(log_entry.user, "poop")
		self.assertEqual(log_entry.timestamp, datetime.datetime(year=2025, month=3, day=10, hour=17, minute=32, second=54))

		with self.assertRaises(exceptions.BinLogParseError):
			BinLogEntry.from_string("Mon Mar 10 17:32:54  Computer:zMichael        User: poop           ")
		with self.assertRaises(exceptions.BinLogParseError):
			BinLogEntry.from_string("Mon Mar 32 17:32:54  Computer: zMichael        User: poop           ")
		
		# Leap year nonsense
		leaplog = BinLogEntry.from_string("Mon Feb 29 17:32:54  Computer: zMichael        User: poop           ", max_year=2025)
		self.assertEqual(leaplog.timestamp.year, 2016)
		
		with self.assertRaises(exceptions.BinLogParseError):
			BinLogEntry.from_string("Tue Feb 29 17:32:54  Computer: zMichael        User: poop           ", max_year=2025)
		

	def test_to_string(self):

		self.assertEqual(EXAMPLE_ENTRY.to_string(), EXAMPLE_STRING)

	def test_from_string_validation(self):

		with self.assertRaises(exceptions.BinLogInvalidFieldError):
			BinLogEntry(timestamp=None)
		with self.assertRaises(exceptions.BinLogInvalidFieldError):
			BinLogEntry(timestamp="Today")

		# Computer
		with self.assertRaises(exceptions.BinLogFieldLengthError):
			BinLogEntry(computer="")
		with self.assertRaises(exceptions.BinLogFieldLengthError):
			BinLogEntry(computer="ReallyBigOldName")
		with self.assertRaises(exceptions.BinLogInvalidFieldError):
			BinLogEntry(computer="My\nGoodness")

		# User
		with self.assertRaises(exceptions.BinLogFieldLengthError):
			BinLogEntry(user="")
		with self.assertRaises(exceptions.BinLogFieldLengthError):
			BinLogEntry(user="ReallyBigOldName")
		with self.assertRaises(exceptions.BinLogInvalidFieldError):
			BinLogEntry(user="My\nGoodness")
	
	def test_copy(self):
		
		self.assertEqual(
			EXAMPLE_ENTRY.copy_with(user="peepee"),
			BinLogEntry(timestamp=EXAMPLE_ENTRY.timestamp, user="peepee", computer=EXAMPLE_ENTRY.computer)
		)

		with self.assertRaises(exceptions.BinLogInvalidFieldError):
			EXAMPLE_ENTRY.copy_with(user="pee\tpee")
	
	def test_sort(self):

		# Cuz those timestamps be different
		self.assertLess(BinLogEntry(), BinLogEntry())
		self.assertNotEqual(BinLogEntry(), BinLogEntry())
		self.assertFalse(BinLogEntry() > BinLogEntry())

if __name__ == "__main__":

	unittest.main()


"""
:mod:`binhistory` is for working with Avid bin ``.log`` files.

It is designed to help developers read and write valid bin logs 
without a whole lotta fuss.

Tenderly over-engineered by Michael Jordan <michael@glowingpixel.com> while he struggled to find his next AE gig.
"""

from . import exceptions, defaults
from ._binlog import BinLog
from ._binlogentry import BinLogEntry


"""
`BinLog` class (a.k.a THE MEAT)
"""

import collections, datetime, typing

from ._binlogentry import BinLogEntry
from .defaults import MAX_ENTRIES, DEFAULT_FILE_EXTENSION
from .exceptions import BinLogTypeError, BinLogNotFoundError, BinNotFoundError

class BinLog(collections.UserList):
	"""An .avb access log"""

	def __init__(self, entries:typing.Optional[typing.Iterable[BinLogEntry]]=None):

		if entries is None:
			super().__init__()
			return

		try:
			entries = list(entries)
		except TypeError as e:
			raise BinLogTypeError(f"`BinLog` must be initialized with an iterable of `BinLogEntry`s, or `None` (got {type(entries).__name__})") from e

		for entry in entries:
			self._validate_item(entry)

		super().__init__(entries)
	
	# Validators
	@staticmethod
	def _validate_item(item:typing.Any):
		"""Validate an item is the proper type"""
		if not isinstance(item, BinLogEntry):
			raise BinLogTypeError(f"Entries must be of type `BinLogEntry` (got {type(item).__name__})")
	
	def __iter__(self) -> typing.Iterator[BinLogEntry]:
		# For typehints
		return super().__iter__()
	
	def __getitem__(self, key:int) -> BinLogEntry:
		# For typehints
		return super().__getitem__(key)
	
	def __setitem__(self, index:int, item:typing.Any):
		self._validate_item(item)
		super().__setitem__(index, item)
	
	def __add__(self, other):
		if not isinstance(other, self.__class__):
			raise BinLogTypeError(f"Can only add another {self.__class__.__name__} (got {type(other).__name__})")
		return super().__add__(other)
	
	def __iadd__(self, other):
		if not isinstance(other, self.__class__):
			raise BinLogTypeError(f"Can only add another {self.__class__.__name__} (got {type(other).__name__})")
		return super().__iadd__(other)
	
	def insert(self, i, item):
		self._validate_item(item)
		return super().insert(i, item)
	
	def append(self, item):
		self._validate_item(item)
		return super().append(item)
	
	def extend(self, other):
		try:
			for o in other:
				self._validate_item(o)
		except TypeError as e:
			raise BinLogTypeError(e) from e
		return super().extend(other)
	
	# Formatters
	def to_string(self) -> str:
		"""Format as string"""
		sorted_entries = sorted(self)[-MAX_ENTRIES:]
		return str().join(e.to_string() + "\n" for e in sorted_entries)

	# Readers
	@classmethod
	def from_bin(cls, bin_path:str, missing_bin_ok:bool=True, max_year:typing.Optional[int]=None) -> "BinLog":
		"""Load an existing .log file for a given bin"""
		return cls.from_path(BinLog.log_path_from_bin_path(bin_path, missing_bin_ok=missing_bin_ok), max_year)

	@classmethod
	def from_path(cls, log_path:str, max_year:typing.Optional[int]=None) -> "BinLog":
		"""Load from an existing .log file"""
		# NOTE: Encountered mac_roman, need to deal with older encodings sometimes

		try:
			with open (log_path, "r") as log_handle:
				return cls.from_stream(log_handle, max_year=max_year)
		except FileNotFoundError as e:
			raise BinLogNotFoundError(f"A log file was not found at the given path {log_path}")
	
	@classmethod
	def from_stream(cls, file_handle:typing.TextIO, max_year:typing.Optional[int]=None) -> "BinLog":
		"""Parse a log from an open file handle"""
		import os
		
		# If we didn't get a `max_year` anywhere else, use the mtime
		if not max_year:
			stat_info = os.fstat(file_handle.fileno())
			max_year =datetime.datetime.fromtimestamp(stat_info.st_mtime).year

		entries = []
		for entry in file_handle:
			entries.append(BinLogEntry.from_string(entry, max_year=max_year))
		
		return cls(entries)

	# Writers
	def to_bin(self, bin_path:str, missing_bin_ok:bool=True):
		"""Write to a log for a given bin"""
		self.to_path(BinLog.log_path_from_bin_path(bin_path, missing_bin_ok=missing_bin_ok))

	def to_path(self, file_path:str):
		"""Write log to filepath"""
		with open(file_path, "w", encoding="utf-8") as output_handle:
			self.to_stream(output_handle)
	
	def to_stream(self, file_handle:typing.TextIO):
		"""Write log to given stream"""
		file_handle.write(self.to_string())

	# Convenience methods	
	def earliest_entry(self) -> typing.Optional[BinLogEntry]:
		"""Get the first/earliest entry from a bin log"""
		return min(self) if self else None

	def latest_entry(self) -> typing.Optional[BinLogEntry]:
		"""Get the last/latest/most recent entry from a bin log"""
		return max(self) if self else None
	
	def users(self) -> typing.List[str]:
		"""Get a list of unique users in the log"""
		return list(set(e.user for e in self))
	
	def computers(self) -> typing.List[str]:
		"""Get a list of unique computers in the log"""
		return list(set(e.computer for e in self))
	
	def timestamps(self) -> typing.List[datetime.datetime]:
		"""Get a list of unique timestamps in the log"""
		return list(set(e.timestamp for e in self))
	
	@classmethod
	def touch(cls, log_path:str, entry:typing.Optional[BinLogEntry]=None):
		"""Add an entry to a log file"""
		import pathlib

		entries = [entry or BinLogEntry()]

		# Read in any existing entries
		if pathlib.Path(log_path).is_file():
			entries.extend(cls.from_path(log_path))
		
		BinLog(entries).to_path(log_path)
	
	@classmethod
	def touch_bin(cls, bin_path:str, entry:typing.Optional[BinLogEntry]=None, missing_bin_ok:bool=True):
		"""Add an entry to a log file for a given bin"""
		cls.touch(BinLog.log_path_from_bin_path(bin_path, missing_bin_ok), entry)
	
	@staticmethod
	def log_path_from_bin_path(bin_path:str, missing_bin_ok:bool=True) -> str:
		"""Determine the expected log path for a given bin path"""
		import pathlib
		if not missing_bin_ok and not pathlib.Path(bin_path).is_file():
			raise BinNotFoundError(f"An existing bin was not found at {bin_path}")
		return str(pathlib.Path(bin_path).with_suffix(DEFAULT_FILE_EXTENSION))

	def __repr__(self) -> str:
		last_entry = self.latest_entry()
		last_entry_str = last_entry.to_string().rstrip() if last_entry else None
		return f"<{self.__class__.__name__} entries={len(self)} last_entry={last_entry_str}>"



import unittest, tempfile, pathlib, time, datetime
from binhistory import BinLog, BinLogEntry, exceptions, defaults

PATH_BIN = str(pathlib.Path(__file__).with_name("example.avb"))
PATH_LOG = str(pathlib.Path(__file__).with_name("example.log"))

class TestBinLog(unittest.TestCase):

	def test_defaults(self):

		default_log = BinLog()
		self.assertFalse(default_log)
		self.assertCountEqual(default_log, [])
	
	def test_create(self):

		with self.assertRaises(exceptions.BinLogTypeError):
			BinLog(BinLogEntry())
		
		with self.assertRaises(exceptions.BinLogTypeError):
			BinLog("peepee")
		
		with self.assertRaises(exceptions.BinLogTypeError):
			BinLog([BinLog(), "uh oh stinky"])
		
		logs = [BinLogEntry() for _ in range(20)] # So BinLogEntry()s have different timestamps
		self.assertCountEqual(BinLog(logs), logs)
		self.assertCountEqual(BinLog(l for l in logs), logs)
		self.assertEqual(len(BinLog(logs)), len(logs))

	def test_from_bin(self):

		self.assertEqual(BinLog.log_path_from_bin_path(PATH_BIN), PATH_LOG)

		log_raw    = pathlib.Path(PATH_LOG).read_text(encoding="utf-8")
		log_parsed = BinLog.from_bin(PATH_BIN)
		self.assertEqual(log_parsed.to_string(), log_raw)

		log_parsed.append(log_parsed.earliest_entry().copy_with(computer="Heehee"))
		self.assertEqual(log_parsed.to_string(), log_raw)
	
	def test_from_bin_missing(self):

		with self.assertRaises(exceptions.BinNotFoundError):
			BinLog.from_bin("example2.avb", missing_bin_ok=False)
		
		with self.assertRaises(exceptions.BinLogNotFoundError):
			BinLog.from_bin("example2.avb", missing_bin_ok=True)

	def test_list_operations(self):

		log = BinLog.from_path(PATH_LOG)
		
		# Check types
		with self.assertRaises(exceptions.BinLogTypeError):
			log.append([BinLogEntry()])
		with self.assertRaises(exceptions.BinLogTypeError):
			log.extend(BinLogEntry())
		with self.assertRaises(exceptions.BinLogTypeError):
			log + BinLogEntry()
		with self.assertRaises(exceptions.BinLogTypeError):
			log[4] = "No"
		
		# Basics
		log.append(BinLogEntry())
		log.extend([BinLogEntry(), BinLogEntry()])
		del log[3]

		# Add logs
		self.assertEqual(len(log), 12)
		log += log
		self.assertEqual(len(log), 12*2)

		# Contains
		entry = BinLogEntry()
		log[4] = entry
		self.assertTrue(entry in log)
		self.assertEqual(log.count(entry), 1)
	
	def test_stats(self):

		log = BinLog.from_path(PATH_LOG)
	
		# Raw strings
		with open(PATH_LOG) as handle_log:
			raw_log = handle_log.read()
		earliest_string = raw_log.split("\n")[0]
		latest_string   = raw_log.split("\n")[-2]	# Very last line is empty

		self.assertEqual(log.earliest_entry().to_string(), earliest_string)
		self.assertEqual(log.latest_entry().to_string(), latest_string)

		# Hacky but DYNAMIC
		user_count = 0
		computer_count = 0
		for user in log.users():
			user_count += raw_log.count(user)
		for computer in log.computers():
			computer_count += raw_log.count(computer)
		
		self.assertEqual(len(log), 10)
		self.assertEqual(user_count, 10)
		self.assertEqual(computer_count, 10)
		self.assertEqual(len(log.timestamps()), 10)
	
	def test_writes(self):

		# From Existing
		existing_log = BinLog.from_path(PATH_LOG)

		with tempfile.TemporaryDirectory() as temp_dir:
			temp_dir = pathlib.Path(temp_dir)

			existing_log.to_path(temp_dir/"from_existing.log")
		
			# Overwite
			existing_log.to_path(temp_dir/"from_existing.log")

			# For bin
			existing_log.to_bin(temp_dir/"poopoo.avb")
			self.assertEqual(BinLog.from_path(temp_dir/"poopoo.log").to_string(), existing_log.to_string())

			# Fail non-existing bin
			with self.assertRaises(exceptions.BinNotFoundError):
				existing_log.to_bin(temp_dir/"poopoo.avb", missing_bin_ok=False)
	
			# Test MAX_ENTRIES by adding an additional and making sure they're rotated properly
			max_log = BinLog.from_path(PATH_LOG)
			curr_timestamp = datetime.datetime.now().replace(microsecond=0)
			max_log.append(BinLogEntry(timestamp=curr_timestamp))			
			self.assertEqual(len(max_log), defaults.MAX_ENTRIES+1)
			max_log.to_path(temp_dir/"overflow.log")

			max_log_check = BinLog.from_path(temp_dir/"overflow.log")
			self.assertEqual(len(max_log_check), defaults.MAX_ENTRIES)
			self.assertEqual(max_log_check.latest_entry().timestamp, curr_timestamp)
			self.assertEqual(max_log_check.earliest_entry().timestamp, max_log[1].timestamp)
	
	def test_touch(self):

		hostname = defaults.DEFAULT_COMPUTER
		username = defaults.DEFAULT_USER

		with tempfile.TemporaryDirectory() as temp_dir:
			temp_dir = pathlib.Path(temp_dir)

			with self.assertRaises(exceptions.BinNotFoundError):
				BinLog.touch_bin(temp_dir/"weewee.avb", missing_bin_ok=False)

			# Allow no bin
			BinLog.touch_bin(temp_dir/"peepee.avb")

			that_new_log = BinLog.from_path(temp_dir/"peepee.log")

			self.assertEqual(len(that_new_log), 1)
			self.assertEqual(that_new_log[0].user, username)
			self.assertEqual(that_new_log[0].computer, hostname)

			first_timestamp = that_new_log[0].timestamp

			time.sleep(1)

			# Touch it again
			BinLog.touch_bin(temp_dir/"peepee.avb")
			that_new_log = BinLog.from_path(temp_dir/"peepee.log")

			self.assertEqual(len(that_new_log), 2)
			self.assertEqual(that_new_log[1].user, username)
			self.assertEqual(that_new_log[1].computer, hostname)

			self.assertEqual(that_new_log[0].timestamp, first_timestamp)
			self.assertNotEqual(that_new_log[0], that_new_log[1])

if __name__ == "__main__":

	unittest.main()


import dataclasses, datetime, typing
from .defaults import DEFAULT_COMPUTER, DEFAULT_USER, FIELD_START_USER, FIELD_START_COMPUTER, DATETIME_STRING_FORMAT, MAX_FIELD_LENGTH
from .exceptions import BinLogParseError, BinLogInvalidFieldError, BinLogFieldLengthError

@dataclasses.dataclass(frozen=True, order=True)
class BinLogEntry:
	"""An entry in a bin log"""

	timestamp:datetime.datetime = dataclasses.field(default_factory=lambda: datetime.datetime.now())
	"""Timestamp of last access"""

	computer:str = DEFAULT_COMPUTER
	"""Hostname of the system which accessed the bin"""

	user:str = DEFAULT_USER
	"""User profile which accessed the bin"""

	def __post_init__(self):
		"""Validate fields"""

		# `user` should be a non-empty string of printable (not \n etc) characters
		if not isinstance(self.user, str):
			raise BinLogInvalidFieldError(f"`user` field must be a string (got {repr(self.user)})")
		elif not self.user.strip() or len(self.user) > MAX_FIELD_LENGTH:
			raise BinLogFieldLengthError(f"`user` field must be between 1 and {MAX_FIELD_LENGTH} characters long (got {len(self.user)})")
		elif not self.user.isprintable():
			raise BinLogInvalidFieldError(f"`user` field contains invalid characters")

		# `computer` should be a non-empty string of printable (not \n etc) characters
		if not isinstance(self.computer, str):
			raise BinLogInvalidFieldError(f"`computer` field must be a string (got {repr(self.computer)})")
		elif not self.computer.strip() or len(self.computer) > MAX_FIELD_LENGTH:
			raise BinLogFieldLengthError(f"`computer` field must be between 1 and {MAX_FIELD_LENGTH} characters long (got {len(self.computer)})")
		elif not self.computer.isprintable():
			raise BinLogInvalidFieldError(f"`computer` field contains invalid characters")
		
		# `timestamp` should be a `datetime.datetime`
		if not isinstance(self.timestamp, datetime.datetime):
			raise BinLogInvalidFieldError(f"`timestamp` field must be a valid `datetime.datetime` object (got {repr(self.timestamp)})")

	def copy_with(self, **fields):
		"""Creates a copy of this `BinLogEntry`, with optional changes to specified fields"""
		from dataclasses import replace
		return replace(self, **fields)
	
	def to_string(self) -> str:
		"""Format the bin log entry as a string"""
		format_datetime       = self.timestamp.strftime(DATETIME_STRING_FORMAT)
		format_entry_computer = FIELD_START_COMPUTER + self.computer
		format_entry_user     = FIELD_START_USER + self.user

		return str().join([
			format_datetime.ljust(21),
			format_entry_computer.ljust(26),
			format_entry_user.ljust(21)
		])
	
	@classmethod
	def from_string(cls, log_entry:str, max_year:typing.Optional[int]=None) -> "BinLogEntry":
		"""Return the log entry from a given log entry string"""

		try:
			entry_datetime   = log_entry[0:19]
			parsed_timestamp = cls._datetime_from_log_timestamp(entry_datetime, max_year)
		except ValueError as e:
			raise BinLogParseError(f"Unexpected value encountered while parsing access time \"{entry_datetime}\" (Assuming a max year of {max_year}): {e}") from e
		
		# Computer name: Observed to be AT LEAST 15 characters.  Likely the max but need to check.
		entry_computer = log_entry[21:47]
		if not entry_computer.startswith(FIELD_START_COMPUTER):
			raise BinLogParseError(f"Unexpected value encountered while parsing computer name: \"{entry_computer}\"")
		parsed_computer = entry_computer[10:].rstrip()

		# User name: Observed to be max 15 characters (to end of line)
		entry_user = log_entry[47:68]
		if not entry_user.startswith(FIELD_START_USER):
			raise BinLogParseError(f"Unexpected value encountered while parsing user name: \"{entry_user}\"")
		parsed_user = entry_user[6:].rstrip()

		return cls(
			timestamp = parsed_timestamp,
			computer  = parsed_computer,
			user      = parsed_user
		)
	
	@staticmethod
	def _datetime_from_log_timestamp(timestamp:str, max_year:typing.Optional[int]=None) -> datetime.datetime:
		"""
		Form a datetime from a given timestamp string
		
		This gets a little complicated  because timestamps in the .log file don't indicate the year, but they DO
		indicate the day of the week.  So, to get a useful :class:``datetime.datetime`` object out of this, "we" need to determine 
		which year the month/day occured on the particular day of the week using ``max_year`` as a starting point 
		(likely a file modified date, or current year), and counting backwards until we get a good match.

		Also accounting for Feb 29 leap year stuff.  It's been fun.
		"""
		
		import calendar

		if max_year is None:
			max_year = datetime.datetime.now().year

		# Account for leap year
		needs_leapyear = "Feb 29" in timestamp
		while needs_leapyear and not calendar.isleap(max_year):
			max_year -= 1

		# Make the initial datetime from known info
		# NOTE: Appending `max_year` here primarily to avoid invalid leap year timestamps
		initial_date = datetime.datetime.strptime(timestamp + " " + str(max_year), DATETIME_STRING_FORMAT + " %Y")

		# Also get the weekday from the timestamp string to compare against the parsed datetime.datetime weekday
		wkday = timestamp[:3]

		# Search backwards up to 11 years (when weekday/date pairs start repeating)
		for year in range(max_year, max_year - 11, -1):

			if needs_leapyear and not calendar.isleap(year):
				continue

			test_date = initial_date.replace(year=year)
			
			if test_date.strftime("%a") == wkday:
				return test_date

		raise ValueError(f"Could not determine a valid year for which {initial_date.month}/{initial_date.day} occurs on a {wkday}")


"""
Bespoke ``Exception`` classes raised by ``pybinhistory``
"""

class BinLogParseError(ValueError):
	"""An invalid value was encountered while parsing the log"""

class BinLogFieldLengthError(ValueError):
	"""A log field is not a valid length (between 1 and ``MAX_FIELD_LENGTH`` chars)"""

class BinLogInvalidFieldError(ValueError):
	"""A log field contains invalid data"""

class BinLogTypeError(TypeError):
	"""A log entry is not a valid type"""

class BinLogNotFoundError(FileNotFoundError):
	"""A log was not found at the given file path"""

class BinNotFoundError(FileNotFoundError):
	"""A bin was not found at the given file path"""


"""
Sane defaults for optimal operation.  Change these at your own risk!
"""

def _get_default_user() -> str:
	import getpass
	try:
		return getpass.getuser() or "pybinhistory"
	except:
		return "pybinhistory"

def _get_default_computer() -> str:
	import socket
	try:
		return socket.gethostname() or "pybinhistory"
	except:
		return "pybinhistory"

DEFAULT_FILE_EXTENSION:str = ".log"
"""The expected file extension for bin log files"""

MAX_ENTRIES:int = 10
"""Maximum log entries allowed in a file"""

MAX_FIELD_LENGTH:int = 15
"""Max number of characters in User or Computer fields"""
# NerdNote: I feel like this comes from NetBIOS max length of 15?

DATETIME_STRING_FORMAT:str = "%a %b %d %H:%M:%S"
"""Datetime string format for bin log entry (Example: Wed Dec 15 09:47:51)"""

FIELD_START_USER:str       = "User: "
DEFAULT_USER:str           = _get_default_user()[:MAX_FIELD_LENGTH]
"""
Default User Profile name to use during the creation of new :py:doc:`binhistory.BinLogEntry` object.

This defaults to the user account that is currently running the python script.
"""

FIELD_START_COMPUTER:str   = "Computer: "
DEFAULT_COMPUTER:str       = _get_default_computer()[:MAX_FIELD_LENGTH]
"""
Default Computer name to use during the creation of new :py:doc:`binhistory.BinLogEntry` object.

This defaults to the hostname of the machine that is currently running the python script.
"""